/*
 * The Rubik's cube.
 *
 * Sed - april 1999.
 *
 * This program is in the public domain.
 *--------------------
 * The main file.
 */

#include <stdio.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "screen.h"
#include "cube.h"
#include "event.h"
#include "device.h"

device d;

char *the_screen;

double EDGE;
char *FONT;
int HELP_X;
int HELP_Y;

int SCREEN_X;
int SCREEN_Y;
double CUBE_SIZE;
double PERSPECTIVE;

char help_message[] = 
"rubix\n\n"
"        by Sed - april 1999 (sed@free.fr)\n"
"        and Jean-Pierre Demailly\n\n"
"Usage : rubix [-i(nit)] [-m(usic)] [-v(ersion)] [-h(elp)]\n"
"              [-font <spec>] [-width w] [-height h] [-edge <dim>]\n"
"              [-load <saved cube>]\n\n"
"Commands within the game :\n"
"        Move the mouse to rotate the cube faces.\n"
"        Faces to be rotated are marked with a cross.\n"
"        Press <Left arrow> or first mouse button to rotate counterclockwise\n"
"        Press <Right arrow> or third mouse button to rotate clockwise\n"
"        Press <Space> to pause/restart the game\n"
"        <h> to show this help message\n"
"        <i> to reinitialize the cube\n"
"        <p> to produce a random permutation of the cube\n"
"        <r> to reset the cube at its home position\n"
"        <s> to save the game to the file \"./rubix.dat\"\n"
"        <+> to shorten perspective distance\n"
"        <-> to switch to imaginary perspective\n"
"        <q> or <Esc> to exit\n\n"
"homepage: http://sed.free.fr/rubix/index.html\n"
"May the penguin be with you.\n\n";

int help_numlines;
int help_maxlen;
char **help_index;

char help1[] = "<Space>  Start/Pause";
char help2[] = "h, H     Help";
 
void usage()
{
  int i;
   
  fprintf(stderr, "\n");
  for (i=0; i<help_numlines; i++)
      fprintf(stderr, "%s\n", help_index[i]);
}

int main(int nb, char **argv)
{
  CUBE cube;
  SCREEN screen;
   
  int i;
  int rand=1;
  char *filename = NULL;
  char *ptr;
   
  FONT = "7x13";
  EDGE = 0.475;

  SCREEN_X = 320;
  SCREEN_Y = 360;
  PERSPECTIVE = 0.2;
   
  ptr = help_message;
  help_numlines = 1;
  i = 0;
  help_index = (char **) malloc(sizeof(char *));
  help_index[0] = ptr;
   
  while(*ptr) {
    if (*ptr=='\n') {
      *ptr = '\0';
      help_index = (char **) realloc(help_index,
			             (help_numlines+1)*sizeof(char *));       
      help_index[help_numlines] = ptr+1;
      ++help_numlines;
      if (i>help_maxlen) help_maxlen = i;
      i = 0;
    }
    ++ptr;
    ++i;
  }
	
  the_screen=&screen.buffer[0];
  d.buffer=screen.buffer;
  screen.buffer = NULL;
  screen.im = 0;
   
  for (i=1; i<nb; i++) {
    if (!strcasecmp(argv[i], "-h") || !strcasecmp(argv[i], "--h") || !strcasecmp(argv[i], "-help") ||
	!strcasecmp(argv[i], "--help")) {
      usage();
      return 0;
    }
    if (!strcasecmp(argv[i], "-i") || !strcasecmp(argv[i], "--i") || !strcasecmp(argv[i], "-init") || !strcasecmp(argv[i], "--init"))
      rand = 0;
    if (!strcasecmp(argv[i], "-v") || !strcasecmp(argv[i], "--v") || !strcasecmp(argv[i], "-version") ||
	!strcasecmp(argv[i], "--version")) {
      fprintf(stderr, "1.0.3\n");
      return 0;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-e") ||
		   !strcasecmp(argv[i], "--e") ||
		   !strcasecmp(argv[i], "-edge") ||
		   !strcasecmp(argv[i], "--edge"))) {
      EDGE = atof(argv[++i]);
      if (EDGE<0.1) EDGE = 0.1;
      if (EDGE>0.5) EDGE = 0.5;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-l") ||
		   !strcasecmp(argv[i], "--l") ||
		   !strcasecmp(argv[i], "-load") ||
		   !strcasecmp(argv[i], "--load"))) {
      filename = argv[++i];
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-fn") ||
		   !strcasecmp(argv[i], "--fn") ||
		   !strcasecmp(argv[i], "-font") ||
		   !strcasecmp(argv[i], "--font"))) {
      FONT = argv[++i];
    }     
    if (i<nb-1 && (!strcasecmp(argv[i], "-width") || 
		   !strcasecmp(argv[i], "--width"))) {
      SCREEN_X = atoi(argv[++i]);
      if (SCREEN_X<80) SCREEN_X = 80;
      if (SCREEN_X>1600) SCREEN_X = 1600;
    }
    if (i<nb-1 && (!strcasecmp(argv[i], "-height") ||
		   !strcasecmp(argv[i], "--height"))) {
      SCREEN_Y = atoi(argv[++i]);
      if (SCREEN_Y<60) SCREEN_Y = 60;
      if (SCREEN_Y>1200) SCREEN_Y = 1200;
    }     
  }
  set_cube_size();
   
  if (init_screen(&screen)==-1) {
    fprintf(stderr, "Error with initing the screen. Are you under X with a 8 bpp resolution ?\n");
    return -1;
  }

  if (filename)
    init_cube(&cube, filename, 0);
  else
    init_cube(&cube, NULL, rand);

  screen.pause = 1;   
  grab_pointer(&screen, 0);
  display_cube(&cube, &screen);

  while(1) {   
    int d=wait_event(&screen, &cube);
     
    if (!(d&SCREEN_EVENT)) {
      if (cube.anim)
	display_cube(&cube, &screen);
    } else {
      if (screen_event(&screen, &cube)==-1)
	break;
    }
  }

  close_screen(&screen);

  return 0;
}
