/*
 * The Rubik's cube.
 *
 * Sed - april 1999.
 *
 * This program is in the public domain.
 *--------------------
 * The screen stuff, ie X.
 * 
 * Heavily hacked by J.-P. Demailly during week-end in October 2003,
 * to make following additions work :
 * wm_delete ClientMessage intercepted + menu popup + automatic resize
 * 
 */

#include "screen.h"

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <X11/cursorfont.h>
#include <string.h>
#include <math.h>
#include <sys/time.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>

#include "cube.h"
#include "device.h"

extern char help_message[];

static Atom wm_protocols, wm_delete_window;

/* original colors designed by Sed
   static int pal[7][3]={{0, 0, 0},
			{63, 0, 0}, 
			{0, 63, 0}, 
			{20, 40, 63}, 
			{63, 63, 0}, 
			{176/4, 48/4, 96/4}, 
			{63, 63, 63}}; */

/* J.-P. Demailly's colors... */
static int pal[7][3]={
     {15, 15, 15},   /* internal faces */
     {63,  0, 0}, 
     {0,  63, 0}, 
     {15, 31, 63}, 
     {63, 63, 0}, 
     {63, 36, 0}, 
     {63, 63, 52}
};

static int _8_16[7];
static int _8_32[7];

void set_colormap(SCREEN *s)
{
  XColor color[256];
  int i;

  if (s->depth!=8)
    return;

  for (i=0; i<7; i++) {
    color[i].pixel=i;
    color[i].flags=DoRed|DoGreen|DoBlue;
    color[i].red=((int)pal[i][0])<<10;
    color[i].green=((int)pal[i][1])<<10;
    color[i].blue=((int)pal[i][2])<<10;
  }
  XStoreColors(s->d, s->cm, color, 7);
}

int to16(int col)
{
  return _8_16[col];
}

int to32(int col)
{
  return _8_32[col];
}

void error_statement(SCREEN *s)
{
  perror("");
  fprintf(stderr, "Meaning no rubix\n");
  XCloseDisplay(s->d);
  s->d=(Display *)0;   
}

int reset_data_buffers(SCREEN *s)
{     

  if (s->im)
    XDestroyImage(s->im);

  s->buffer=(char *)malloc(((SCREEN_X+7)/8)*SCREEN_Y*s->depth);

  if (!s->buffer) {
    error_statement(s);
    return -1;
  }
   
  if (the_lines) 
    free(the_lines);
  the_lines = (device_line *)malloc(SCREEN_Y*sizeof(device_line));   

  if (!the_lines) {
    error_statement(s);
    return -1;     
  }   

  /* create our XImage */
  /* bitmap_pad, what the heck is it ?? mail sed@free.fr if you know, please...
   * Well, seems it has to be 32...
   */
  s->im=XCreateImage(s->d, DefaultVisual(s->d, DefaultScreen(s->d)), 
		     s->depth==32?24:s->depth, ZPixmap,
		     0, s->buffer, SCREEN_X, SCREEN_Y, 32, 0);

  if (!s->im) {
    error_statement(s);
    return -1;     
  } else
    return 0;
}

void set_cube_size()
{
  int s;
  s = SCREEN_X;
  if (SCREEN_Y < s) s = SCREEN_Y;
  CUBE_SIZE = s * (0.95 - 0.2 * fabs(PERSPECTIVE));
}

/* return -1 if error, 0 if ok */
int init_screen(SCREEN *s)
{
  extern device d;
  XSetWindowAttributes attributs;
  XGCValues gcv;
  XEvent ev;
  int i;

  for (i=0; i<7; i++) {
    _8_16[i]=((pal[i][0]/2)<<(5+6))+(pal[i][1]<<5)+(pal[i][2]/2);;
  }

  for (i=0; i<7; i++) {
    int r, g, b;
    r=(double)pal[i][0]/63.*255.;
    g=(double)pal[i][1]/63.*255.;
    b=(double)pal[i][2]/63.*255.;
    _8_32[i]=(r<<16)|(g<<8)|b;
  }

  memset(s, 0, sizeof(SCREEN));

  if (!(s->d=XOpenDisplay((char *)0)))
    return -1;

  s->depth=DefaultDepth(s->d, DefaultScreen(s->d));
  
  if (s->depth!=8 && s->depth!=16 && s->depth!=24) {
    fprintf(stderr, "screen depth not supported (only 8, 16 and 24bpp (which means 32bpp too) handled\n");
    XCloseDisplay(s->d);
    s->d=(Display *)0;
    return -1;
  }
  
  wm_protocols = XInternAtom(s->d, "WM_PROTOCOLS", False);
  wm_delete_window = XInternAtom(s->d, "WM_DELETE_WINDOW", False);
   
  s->font = XLoadQueryFont(s->d, FONT);
  if (!s->font) {
    fprintf(stderr, "cannot load font %s !!\n", FONT);
    return -1;
  }
  HELP_X = (help_maxlen + 1) * XTextWidth(s->font, "w", 1);
  HELP_Y = (help_numlines + 2) *
           (s->font->max_bounds.ascent + s->font->max_bounds.descent + 2);
   
  /* we don't know if s->depth==24 if it's 24 or 32bpp, let's check it */
  if (s->depth==24) {
    int n;
    XPixmapFormatValues *v;

    v=XListPixmapFormats(s->d, &n);
    if (!v || !n) {
      if (v)
	XFree(v);
      fprintf(stderr, "Error while trying to distinguish between 24 or 32bpp, damned...\n");
      XCloseDisplay(s->d);
      s->d=(Display *)0;
      return -1;
    }
    for (n--; n>=0; n--)
      if (v[n].bits_per_pixel==32)
	s->depth=32;
    XFree(v);
  }

  if (s->depth==8) {
    s->cm=XCreateColormap(s->d, DefaultRootWindow(s->d), DefaultVisual(s->d, DefaultScreen(s->d)), AllocAll);
    attributs.colormap=s->cm;
  }
   
  /* let's create and map our window */
  if (s->depth==8) {
    s->w=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 
		       SCREEN_X, SCREEN_Y, 3, 8, CopyFromParent, CopyFromParent,
		       CWColormap, &attributs);
    s->h=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 
		       HELP_X, HELP_Y, 3, 8, CopyFromParent, CopyFromParent,
		       CWColormap, &attributs);     
  } else {
    s->w=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 
		       SCREEN_X, SCREEN_Y, 3, s->depth==32?24:s->depth,
		       CopyFromParent, CopyFromParent, 0, NULL);
    s->h=XCreateWindow(s->d, DefaultRootWindow(s->d), 0, 0, 
		       HELP_X, HELP_Y, 3, s->depth==32?24:s->depth,
		       CopyFromParent, CopyFromParent, 0, NULL);     
  }
   
  {
    XClassHint          xch;
    xch.res_name = "rubix";
    xch.res_class = "Rubix";
    XSetClassHint(s->d, s->w, &xch);
    XStoreName(s->d, s->w, xch.res_name);
    xch.res_name = "rubix/help"; 
    XStoreName(s->d, s->h, xch.res_name);     
  }
   
  XSelectInput(s->d, s->w, 
    KeyPressMask|ExposureMask|StructureNotifyMask|
    PointerMotionMask|ButtonPressMask);
  XSetWMProtocols(s->d, s->w, &wm_delete_window, 1);
   
  XSelectInput(s->d, s->h, ExposureMask|ButtonReleaseMask);
  XSetWMProtocols(s->d, s->h, &wm_delete_window, 1);
   
  XSetWindowBackground(s->d, s->h, 
    BlackPixelOfScreen(DefaultScreenOfDisplay(s->d)));
   
  XMapWindow(s->d, s->w);
  do
    XNextEvent(s->d, &ev);
  while (ev.type!=Expose || ev.xexpose.window!=s->w);
  XFlush(s->d);
  XSetWMProtocols(s->d, s->w, &wm_delete_window, 1);

  gcv.function=GXcopy;
  s->gc=XCreateGC(s->d,s->w,GCFunction,&gcv);
  XSetFont(s->d, s->gc, s->font->fid);
     
  if (reset_data_buffers(s) == -1)
    return -1;

  d.buffer=s->buffer;
  d.depth=s->depth;
   
  if (s->depth==8)
    set_colormap(s);

  return 0;
}

void grab_pointer(SCREEN *s, int mode)
{
  static Cursor void_cursor = None;
  static Cursor arrow_cursor = None;   
  Pixmap p;
  XColor a;
  GC pixgc;
  XSetWindowAttributes attributs;
     
  if (mode == 0) {
    XUngrabPointer(s->d, 0);
    if (arrow_cursor == None)
      arrow_cursor = XCreateFontCursor(s->d, XC_hand2);
    attributs.cursor = arrow_cursor;
    XChangeWindowAttributes(s->d, s->w, CWCursor, &attributs);
    return;
  }
   
	
  /* let's create a void cursor for our window */
  if (void_cursor == None) {
    p=XCreatePixmap(s->d, DefaultRootWindow(s->d), 1, 1, 1);
    pixgc=XCreateGC(s->d, p, 0, 0);
    a.red=a.green=a.blue=0;
    XFillRectangle(s->d, p, pixgc, 0, 0, 2, 2);
    XFreeGC(s->d, pixgc);
    void_cursor = XCreatePixmapCursor(s->d, p, p, &a, &a, 0, 0);
    /* XFreePixmap(s->d, p); */
  }
  attributs.cursor = void_cursor;
  XChangeWindowAttributes(s->d, s->w, CWCursor, &attributs);
   
  if (XGrabPointer(s->d, s->w, False, 
      PointerMotionMask|ButtonPressMask|ButtonReleaseMask,
      GrabModeAsync, GrabModeAsync, s->w, None, 0))
    fprintf(stderr, "Could not grab the pointer, the program won't run very nice.\n");
  place_mouse(s);
}

void display_help(SCREEN *s)
{
  int i, dy;
  XFlush(s->d);   
  usleep(50000);   
  XSetForeground(s->d, s->gc, 
		   WhitePixelOfScreen(DefaultScreenOfDisplay(s->d)));
  dy = s->font->max_bounds.ascent + s->font->max_bounds.descent + 2;
  for (i=0; i<help_numlines; i++)
  XDrawString(s->d, s->h, s->gc, 6,
		s->font->max_bounds.ascent + 6 + i * dy,
		help_index[i], strlen(help_index[i]));
  XDrawString(s->d, s->h, s->gc, HELP_X - 53, 
	      HELP_Y -  s->font->max_bounds.descent - 10, "OK", 2);
  XDrawRectangle(s->d, s->h, s->gc, HELP_X - 80, HELP_Y - dy - 12, 70, dy+4);
  XDrawRectangle(s->d, s->h, s->gc, HELP_X - 81, HELP_Y - dy - 13, 72, dy+6);
  XFlush(s->d);
}

/* return -1 if end, 0 otherwise */
int screen_key(SCREEN *s, CUBE *c, XKeyEvent *ev)
{
  char dum[50];
  KeySym ks;
  int n;

  XLookupString(ev, dum, 50, &ks, (XComposeStatus *)0);

  switch(ks) {
    case XK_Escape :
    case XK_q :
    case XK_Q :
      return -1;
    case XK_plus :
    case XK_KP_Add :
         n = (int) (10*PERSPECTIVE + 10.5);
         ++n;
         if (n>20) n = 20;
         PERSPECTIVE = 0.1 * n - 1.0;
         set_cube_size();
         display_cube(c, s);     
         break;
    case XK_minus :
    case XK_KP_Subtract :
         n = (int) (10*PERSPECTIVE + 10.5);
         --n;
         if (n<0) n = 0;
         PERSPECTIVE = 0.1 * n - 1.0;
         set_cube_size();     
         display_cube(c, s);     
         break;     
    case XK_h :
    case XK_H :
      XMapRaised(s->d, s->h);
      XClearWindow(s->d, s->h);
      display_help(s);
      s->pause = 0;
    case XK_space :
      s->pause=1-s->pause;
      if (s->pause) {
	grab_pointer(s, 0);
      } else {
	grab_pointer(s, 1);
      }
      display_cube(c, s);     
      break;
    case XK_i :
    case XK_I :
      init_cube(c, NULL, -1);
      display_cube(c, s);
      break;
    case XK_p :
    case XK_P :
      init_cube(c, NULL, 1);
      display_cube(c, s);
      break;
    case XK_r :
    case XK_R :
      reset_cube(c, s);
      display_cube(c, s);
      break;
    case XK_s :
    case XK_S :
      save_cube(c);
      break;
    case XK_Left :
      if (!s->pause)
	cube_generate_rotate(c, s, 0);
      break;
    case XK_Right :
      if (!s->pause)
	cube_generate_rotate(c, s, 1);
      break;
  }

  return 0;
}

void place_mouse(SCREEN *s)
{
  if (!s->pause) {
    XWarpPointer(s->d, None, s->w, 0, 0, 0, 0, SCREEN_X/2, SCREEN_Y/2);
    XFlush(s->d);
  }
}

void mouse_pre_event(XMotionEvent *ev, int *a, int *b)
{
  if (ev->x==SCREEN_X/2 && ev->y==SCREEN_Y/2)
    return;

  *a=ev->y-SCREEN_Y/2;
  *b=ev->x-SCREEN_X/2;
}

void mouse_event(SCREEN *s, CUBE *c, int a, int b)
{
#define SPEED (1024.)
  if (s->pause)
    return;

  if (a==0 && b==0)
    return;

  place_mouse(s);

  /* the x is y and vice-versa */
  cube_rotate(s, c, a*M_PI/SPEED, b*M_PI/SPEED);
  display_cube(c, s);
}

void get_window_size(SCREEN *s, int *x, int*y)
{
  Window root, parent, *children;
  int u, v, b, d, n;
  XQueryTree(s->d, s->w, &root, &parent, &children, &n);
   
  if (!parent) {   	
    fprintf(stderr, "Cannot query window tree!\n");
    return;
  }
   
  XGetGeometry(s->d, parent, &root, &u, &v, x, y, &b, &d);
}

Bool evpred(Display *d, XEvent *ev, XPointer a)
{   
  return (True);
}


/* return -1 if end, 0 if ok */
int screen_event(SCREEN *s, CUBE *c)
{
#define EXPOSE 1
#define MOTION 2
  extern device d;
  int type=0;
  int a=0, b=0, dy;
  XEvent ev;

  while(XCheckIfEvent(s->d, &ev, evpred, (XPointer)0)) {
    switch(ev.type) {
      case ClientMessage :
        if (ev.xclient.message_type == wm_protocols &&
          ev.xclient.format == 32 &&
          ev.xclient.data.l[0] == wm_delete_window) {
          if (ev.xexpose.window == s->w) {
            close_screen(s);
            exit(0);
          }
          if (ev.xexpose.window == s->h) {
            XUnmapWindow(s->d, s->h);
          }
	}
        break;
      case Expose :
	type|=EXPOSE;
	break;
      case KeyPress :
	if (screen_key(s, c, &ev.xkey)==-1)
	  return -1;
	break;
      case ButtonRelease :
        dy = s->font->max_bounds.ascent + s->font->max_bounds.descent + 2;        
        if (ev.xexpose.window==s->h &&
	    ev.xbutton.x>=HELP_X - 80 && ev.xbutton.x<=HELP_X - 10 &&
	    ev.xbutton.y>=HELP_Y - dy - 12 && ev.xbutton.y<=HELP_Y - 8)
	   XUnmapWindow(s->d, s->h);
        break;
      case ButtonPress :
        if (ev.xexpose.window != s->w) 
	  break;
	if (ev.xbutton.button==Button1) {
	  cube_generate_rotate(c, s, 0);
	} else if (ev.xbutton.button==Button3) {
	  cube_generate_rotate(c, s, 1);
	}
	break;
      case MotionNotify :
	mouse_pre_event(&ev.xmotion, &a, &b);
	type |= MOTION;
	break;
      case PropertyNotify:
      case ConfigureNotify: {
	  int x, y;
	  get_window_size(s, &x, &y);
	  if (x!=SCREEN_X || y!=SCREEN_Y) {
	    if (x<80) x = 80;
	    if (y<60) y = 60;	     
	    XResizeWindow(s->d, s->w, x, y);
	    XSync(s->d, True);
	    SCREEN_X = x;
	    SCREEN_Y = y;
	    reset_data_buffers(s);
	    d.buffer=s->buffer;
	    set_cube_size();
	    display_cube(c, s);
	    XSync(s->d, True);	   
	  }
        }
        break;
#if 0
     case FocusOut :
        if (!s->pause)
	  place_mouse(s);
	break;
#endif
    }
  }
  if (type & MOTION) {
    mouse_event(s, c, a, b);
  } else if (type & EXPOSE) {
    if (ev.xexpose.window == s->w) put_screen(s);
    if (ev.xexpose.window == s->h) display_help(s);     
  }
  return 0;
}

void close_screen(SCREEN *s)
{
  XDestroyWindow(s->d, s->w);
  XFreeGC(s->d, s->gc);
  if (s->depth==8)
    XFreeColormap(s->d, s->cm);
  XCloseDisplay(s->d);
}

void clear_screen(SCREEN *s)
{
  memset(s->buffer, 0, ((SCREEN_X+7)/8)*SCREEN_Y*s->depth);
}

void put_screen(SCREEN *s)
{
  XPutImage(s->d, s->w, s->gc, s->im, 0, 0, 0, 0, SCREEN_X, SCREEN_Y);
  if (s->pause) {
    XSetForeground(s->d, s->gc, 
		   WhitePixelOfScreen(DefaultScreenOfDisplay(s->d)));
    XDrawString(s->d, s->w, s->gc, 6, 
		s->font->max_bounds.ascent + 6, help1, strlen(help1));
    XDrawString(s->d, s->w, s->gc, 6, 
		SCREEN_Y - s->font->max_bounds.descent - 6,
		help2, strlen(help2));
    XFlush(s->d);
  }
  /* XFlush(s->d); */
}

/* return ms */
struct timeval get_time(void)
{
  struct timeval t;
  struct timezone d;
  if (gettimeofday(&t, &d)==-1) {
    perror("gettimeofday");
    return t;
  }
  return t;
}

